//go:generate mockery --case snake --all

package crypto

import (
	"crypto/aes"
	"crypto/cipher"
	"crypto/sha1"
	"encoding/base64"
	"errors"
	"fmt"
	mrand "math/rand"
	"strings"
	"time"

	"golang.org/x/crypto/pbkdf2"
)

const letterBytes = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

func init() {
	mrand.Seed(time.Now().UnixNano())
}

const (
	separator = ":"
)

var (
	ErrNoVersion        = errors.New("crypto: version not set for SecretsHandler")
	ErrCypherTextParse  = errors.New("crypto: parse cypher text")
	ErrVersionMismatch  = errors.New("crypto: SecretsHandler not configured for specified cypher text")
	ErrNoPassphrase     = errors.New("crypto: SecretsHandler requires a passphrase")
	ErrCypherTextBroken = errors.New("crypto: Cipher text broken")
)

type Crypter interface {
	Encrypt(plaintext string) (string, error)
	Decrypt(ciphertext string) (string, error)
}

// SecretsHandler handles all cryptographic operations in this package
type SecretsHandler struct {
	passphrase string
	version    *Version
	padder     *padder
}

// NewSecretsHandler returns an instance of SecretsHandler that can handle
// cryptographic operations using the specified version and block size
func NewSecretsHandler(passphrase string, version *Version) (*SecretsHandler, error) {
	if passphrase == "" {
		return nil, ErrNoPassphrase
	}
	if version == nil {
		return nil, ErrNoVersion
	}

	p, err := newPadder(version.Padding)
	if err != nil {
		return nil, err
	}

	return &SecretsHandler{
		passphrase: passphrase,
		version:    version,
		padder:     p,
	}, nil
}

// Encrypt encrypts the specified plain text, using the specfied pass phrase,
// and returns the cypher text.
func (sh *SecretsHandler) Encrypt(plaintext string) (string, error) {
	salt, err := randomBytes(8)
	if err != nil {
		return "", err
	}
	iv, err := randomBytes(16)
	if err != nil {
		return "", err
	}

	key := pbkdf2.Key([]byte(sh.passphrase), salt, sh.version.Iterations, sh.version.Bits/8, sha1.New)
	block, err := aes.NewCipher(key)
	if err != nil {
		return "", err
	}

	enctext := sh.padder.pad([]byte(plaintext))

	cbc := cipher.NewCBCEncrypter(block, iv)
	cbc.CryptBlocks(enctext, enctext)

	ciphertext := fmt.Sprintf("%s%s%s%s%s%s%s", sh.version.Name,
		separator, base64.StdEncoding.EncodeToString(salt),
		separator, base64.StdEncoding.EncodeToString(iv),
		separator, base64.StdEncoding.EncodeToString(enctext))

	return ciphertext, nil
}

// Decrypt decrypts the specified cypher text, using the specified pass phrase,
// and returns the plain text.
func (sh *SecretsHandler) Decrypt(ciphertext string) (string, error) {
	version, salt, iv, enctext, err := sh.parse(ciphertext)
	if err != nil {
		return "", err
	}

	if *version != sh.version.Name {
		return "", ErrVersionMismatch
	}

	key := pbkdf2.Key([]byte(sh.passphrase), salt, sh.version.Iterations, sh.version.Bits/8, sha1.New)
	block, err := aes.NewCipher(key)
	if err != nil {
		return "", err
	}
	cbc := cipher.NewCBCDecrypter(block, iv)
	if len(enctext)%cbc.BlockSize() != 0 {
		return "", ErrCypherTextBroken
	}
	cbc.CryptBlocks(enctext, enctext)

	return sh.padder.trim(enctext)
}

func (sh *SecretsHandler) parse(cypherText string) (*string, []byte, []byte, []byte, error) {
	components := strings.Split(cypherText, separator)
	if len(components) != 4 {
		return nil, nil, nil, nil, ErrCypherTextParse
	}

	salt, err := base64.StdEncoding.DecodeString(components[1])
	if err != nil {
		return nil, nil, nil, nil, ErrCypherTextParse
	}
	iv, err := base64.StdEncoding.DecodeString(components[2])
	if err != nil {
		return nil, nil, nil, nil, ErrCypherTextParse
	}
	enctext, err := base64.StdEncoding.DecodeString(components[3])
	if err != nil {
		return nil, nil, nil, nil, ErrCypherTextParse
	}

	return &components[0], salt, iv, enctext, nil
}

// GenerateKey generates API keys (default n=40)
func GenerateKey(n int) string {
	b := make([]byte, n)
	for i := range b {
		b[i] = letterBytes[mrand.Int63()%int64(len(letterBytes))]
	}

	return string(b)
}
