package crypto

import (
	"encoding/hex"
	"errors"
	"strconv"
)

const (
	pkcs7 padding = "PKCS7Padding"
	pkcs5 padding = "PKCS5Padding"

	pkcs7BlockSize = 16
	pkcs5BlockSize = 8
)

var (
	ErrInvalidPadding = errors.New("crypto: unsupported padding")
	ErrInputPadding   = errors.New("crypto: input padding error")
)

type padding string

type padder struct {
	blockSize int
}

func newPadder(p padding) (*padder, error) {
	if p == pkcs7 {
		return &padder{blockSize: pkcs7BlockSize}, nil
	}
	if p == pkcs5 {
		return &padder{blockSize: pkcs5BlockSize}, nil
	}

	return nil, ErrInvalidPadding
}

func (p *padder) pad(text []byte) []byte {
	c := p.blockSize - (len(text) % p.blockSize)
	for i := 0; i < c; i++ {
		text = append(text, byte(c))
	}

	return text
}

func (p *padder) trim(text []byte) (string, error) {
	nl := len(text)
	x := hex.EncodeToString(text[nl-1:])
	c, err := strconv.ParseInt(x, 16, p.blockSize)
	if err != nil {
		return "", err
	}

	if c > int64(p.blockSize) {
		return "", ErrInputPadding
	}

	return string(text[:nl-int(c)]), nil
}
