package logging

import (
	"github.com/prometheus/client_golang/prometheus"
	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"

	"github.com/greenqloud/heimdall/internal/config"
)

var counterErrors = prometheus.NewCounterVec(prometheus.CounterOpts{
	Subsystem: "logging",
	Name:      "errors_total",
	Help:      "Number of entries by log level",
},
	[]string{"level"})

func init() {
	prometheus.MustRegister(counterErrors)
}

type Logger struct {
	*zap.SugaredLogger
}

func FromConfig(cfg Config) (*Logger, error) {
	c := zap.NewProductionConfig()
	if !cfg.Structured {
		c.Encoding = "console"
		c.EncoderConfig = zap.NewDevelopmentEncoderConfig()
	}

	_ = c.Level.UnmarshalText([]byte(cfg.Level))
	c.DisableStacktrace = !cfg.Stacktrace

	// Caller skip allows us to display the original caller in logs, opposed to just displaying our log wrappers
	skip := zap.AddCallerSkip(1)
	l, err := c.Build(skip)
	if err != nil {
		return nil, err
	}

	return &Logger{
		SugaredLogger: l.Sugar(),
	}, nil
}

func New() *Logger {
	var cfg = new(Config)
	_ = config.Process(cfg)
	logger, _ := FromConfig(*cfg)
	return logger
}

func NoOp() *Logger {
	return &Logger{
		SugaredLogger: zap.NewNop().Sugar(),
	}
}

func NewWithContext(context ...interface{}) *Logger {
	return New().With(context)
}

func (l *Logger) With(context ...interface{}) *Logger {
	return &Logger{
		SugaredLogger: l.SugaredLogger.With(context...),
	}
}

func (l *Logger) Error(args ...interface{}) {
	instrument(zapcore.ErrorLevel)
	l.SugaredLogger.Error(args...)
}

func (l *Logger) Errorf(template string, args ...interface{}) {
	instrument(zapcore.ErrorLevel)
	l.SugaredLogger.Errorf(template, args...)
}

func (l *Logger) Errorw(msg string, keysAndValues ...interface{}) {
	instrument(zapcore.ErrorLevel)
	l.SugaredLogger.Errorw(msg, keysAndValues...)
}

func (l *Logger) Warn(args ...interface{}) {
	instrument(zapcore.WarnLevel)
	l.SugaredLogger.Warn(args...)
}

func (l *Logger) Warnf(template string, args ...interface{}) {
	instrument(zapcore.WarnLevel)
	l.SugaredLogger.Warnf(template, args...)
}

func (l *Logger) Warnw(msg string, keysAndValues ...interface{}) {
	instrument(zapcore.WarnLevel)
	l.SugaredLogger.Warnw(msg, keysAndValues...)
}

func instrument(level zapcore.Level) {
	counterErrors.WithLabelValues(level.String()).Inc()
}
