package sign

import (
	"crypto/hmac"
	"crypto/sha1"
	"encoding/base64"
	"errors"
	"fmt"
	"io"
	"net/http"
	"net/url"
	"sort"
	"strings"
)

var (
	ErrFailedToCreateSignature = errors.New("auth: failed to create signature")
)

type Signer interface {
	Sign(r *http.Request) (string, error)
}

type RequestSigner struct {
	SecretKey string
}

func (rs RequestSigner) Sign(r *http.Request) (string, error) {
	params := rs.writeQuery(r)
	msg := r.URL.Path + params

	serverSignature, err := rs.createSignature(msg)
	if err != nil {
		return "", ErrFailedToCreateSignature
	}

	return serverSignature, nil
}

func (rs RequestSigner) writeQuery(r *http.Request) string {
	var params []string
	for k, vs := range r.URL.Query() {
		k = url.QueryEscape(k)
		if k == "signature" {
			continue
		}
		for _, v := range vs {
			if v == "" {
				params = append(params, k)
			} else {
				v = url.QueryEscape(v)
				params = append(params, k+"="+v)
			}
		}
	}
	sort.Strings(params)
	return fmt.Sprintf("?%s", strings.Join(params, "&"))
}

func (rs RequestSigner) createSignature(msg string) (string, error) {
	serverSignature := hmac.New(sha1.New, []byte(rs.SecretKey))
	_, err := io.WriteString(serverSignature, msg)
	if err != nil {
		return "", err
	}
	encodedServerSignature := base64.StdEncoding.EncodeToString(serverSignature.Sum(nil))
	return encodedServerSignature, nil
}

// TrailingSlashSigner is a temporary measure of implementation where we have
// a client that is signing requests with a trailing slash included.
// The trailing slash signer will attempt to sign a request by adding an additional trailing slash.
type TrailingSlashSigner struct {
	RequestSigner
}

func (rs TrailingSlashSigner) Sign(r *http.Request) (string, error) {
	req := r.Clone(r.Context())
	req.URL.Path = req.URL.Path + "/"
	return rs.RequestSigner.Sign(req)
}
